#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "GameplayLoop.h"

// reference to FinishGame (in main.cpp) so we can bail
extern void FinishGame(GlobalData *globals);

/**
 * Initialize gameplay loop state
 */
void CGameplayLoop::Init(GlobalData *inGlobals, char **wordList, int nWords, char *LevelLabel, char **inIntro) {
	int i, j;
	this->globals = inGlobals;	// save pointer to globals
	int minWordLen = 100;
	int remainder = 0;

	// First, let's get 10 lines of randomly generated text:

	// hey it's the worst way to do this but it's fine
	// for picking random words from a short list
	srand(time(NULL));

	// find the shortest word on the list
	for (i=0;i<nWords;i++) {
		if (strlen(wordList[i]) < minWordLen) {
			minWordLen = strlen(wordList[i]);
		}
	}

	// clear strings
	for (i=0;i<10;i++) {
		this->GeneratedLines[i][0] = '\0';
		remainder = 36;

		// while there's still space for at least the smallest word
		while (remainder > minWordLen + 1) {
			// pick a random word
			j = rand()%nWords;
			// if it will fit on the line
			if (strlen(wordList[j]) + 1 < remainder) {
				// if this isn't the first word, put a space first
				if (this->GeneratedLines[i][0] != '\0') {
					strcat(this->GeneratedLines[i], " ");
					remainder --;
				}
				// append the word
				strcat(this->GeneratedLines[i], wordList[j]);
				// subtract its length from the remaining space
				remainder -= strlen(wordList[j]);
			}
		}
	}

	// start building the keyboard of printable characters
	nKeys = 0;
	this->keylist[nKeys++] =		this->DefineKey('`', '~', 0, 0, 2, 2, SDL_SCANCODE_GRAVE, false);
	this->keylist[nKeys++] =		this->DefineKey('1', '!', 2, 0, 2, 2, SDL_SCANCODE_1, false);
	this->keylist[nKeys++] =		this->DefineKey('2', '@', 4, 0, 2, 2, SDL_SCANCODE_2, false);
	this->keylist[nKeys++] =		this->DefineKey('3', '#', 6, 0, 2, 2, SDL_SCANCODE_3, false);
	this->keylist[nKeys++] =		this->DefineKey('4', '$', 8, 0, 2, 2, SDL_SCANCODE_4, false);
	this->keylist[nKeys++] =		this->DefineKey('5', '%', 10, 0, 2, 2, SDL_SCANCODE_5, false);
	this->keylist[nKeys++] =		this->DefineKey('6', '^', 12, 0, 2, 2, SDL_SCANCODE_6, false);
	this->keylist[nKeys++] =		this->DefineKey('7', '&', 14, 0, 2, 2, SDL_SCANCODE_7, false);
	this->keylist[nKeys++] =		this->DefineKey('8', '*', 16, 0, 2, 2, SDL_SCANCODE_8, false);
	this->keylist[nKeys++] =		this->DefineKey('9', '(', 18, 0, 2, 2, SDL_SCANCODE_9, false);
	this->keylist[nKeys++] =		this->DefineKey('0', ')', 20, 0, 2, 2, SDL_SCANCODE_0, false);
	this->keylist[nKeys++] =		this->DefineKey('-', '_', 22, 0, 2, 2, SDL_SCANCODE_MINUS, false);
	this->keylist[nKeys++] =		this->DefineKey('=', '+', 24, 0, 2, 2, SDL_SCANCODE_EQUALS, false);
	
	this->keylist[nKeys++] =		this->DefineKey('q', 'Q', 3, 2, 2, 2, SDL_SCANCODE_Q, true);
	this->keylist[nKeys++] =		this->DefineKey('w', 'W', 5, 2, 2, 2, SDL_SCANCODE_W, true);
	this->keylist[nKeys++] =		this->DefineKey('e', 'E', 7, 2, 2, 2, SDL_SCANCODE_E, true);
	this->keylist[nKeys++] =		this->DefineKey('r', 'R', 9, 2, 2, 2, SDL_SCANCODE_R, true);
	this->keylist[nKeys++] =		this->DefineKey('t', 'T', 11, 2, 2, 2, SDL_SCANCODE_T, true);
	this->keylist[nKeys++] =		this->DefineKey('y', 'Y', 13, 2, 2, 2, SDL_SCANCODE_Y, true);
	this->keylist[nKeys++] =		this->DefineKey('u', 'U', 15, 2, 2, 2, SDL_SCANCODE_U, true);
	this->keylist[nKeys++] =		this->DefineKey('i', 'I', 17, 2, 2, 2, SDL_SCANCODE_I, true);
	this->keylist[nKeys++] =		this->DefineKey('o', 'O', 19, 2, 2, 2, SDL_SCANCODE_O, true);
	this->keylist[nKeys++] =		this->DefineKey('p', 'P', 21, 2, 2, 2, SDL_SCANCODE_P, true);
	this->keylist[nKeys++] =		this->DefineKey('[', '{', 23, 2, 2, 2, SDL_SCANCODE_LEFTBRACKET, false);
	this->keylist[nKeys++] =		this->DefineKey(']', '}', 25, 2, 2, 2, SDL_SCANCODE_RIGHTBRACKET, false);
	this->keylist[nKeys++] =		this->DefineKey('\\', '|', 27, 2, 3, 2, SDL_SCANCODE_BACKSLASH, false);

	this->keylist[nKeys++] =		this->DefineKey('a', 'A', 4, 4, 2, 2, SDL_SCANCODE_A, true);
	this->keylist[nKeys++] =		this->DefineKey('s', 'S', 6, 4, 2, 2, SDL_SCANCODE_S, true);
	this->keylist[nKeys++] =		this->DefineKey('d', 'D', 8, 4, 2, 2, SDL_SCANCODE_D, true);
	this->keylist[nKeys++] =		this->DefineKey('f', 'F', 10, 4, 2, 2, SDL_SCANCODE_F, true);
	this->keylist[nKeys++] =		this->DefineKey('g', 'G', 12, 4, 2, 2, SDL_SCANCODE_G, true);
	this->keylist[nKeys++] =		this->DefineKey('h', 'H', 14, 4, 2, 2, SDL_SCANCODE_H, true);
	this->keylist[nKeys++] =		this->DefineKey('j', 'J', 16, 4, 2, 2, SDL_SCANCODE_J, true);
	this->keylist[nKeys++] =		this->DefineKey('k', 'K', 18, 4, 2, 2, SDL_SCANCODE_K, true);
	this->keylist[nKeys++] =		this->DefineKey('l', 'L', 20, 4, 2, 2, SDL_SCANCODE_L, true);
	this->keylist[nKeys++] =		this->DefineKey(';', ':', 22, 4, 2, 2, SDL_SCANCODE_SEMICOLON, false);
	this->keylist[nKeys++] =		this->DefineKey('\'', '"', 24, 4, 2, 2, SDL_SCANCODE_APOSTROPHE, false);
	
	this->keylist[nKeys++] =		this->DefineKey('z', 'Z', 5, 6, 2, 2, SDL_SCANCODE_Z, true);
	this->keylist[nKeys++] =		this->DefineKey('x', 'X', 7, 6, 2, 2, SDL_SCANCODE_X, true);
	this->keylist[nKeys++] =		this->DefineKey('c', 'C', 9, 6, 2, 2, SDL_SCANCODE_C, true);
	this->keylist[nKeys++] =		this->DefineKey('v', 'V', 11, 6, 2, 2, SDL_SCANCODE_V, true);
	this->keylist[nKeys++] =		this->DefineKey('b', 'B', 13, 6, 2, 2, SDL_SCANCODE_B, true);
	this->keylist[nKeys++] =		this->DefineKey('n', 'N', 15, 6, 2, 2, SDL_SCANCODE_N, true);
	this->keylist[nKeys++] =		this->DefineKey('m', 'M', 17, 6, 2, 2, SDL_SCANCODE_M, true);
	this->keylist[nKeys++] =		this->DefineKey(',', '<', 19, 6, 2, 2, SDL_SCANCODE_COMMA, false);
	this->keylist[nKeys++] =		this->DefineKey('.', '>', 21, 6, 2, 2, SDL_SCANCODE_PERIOD, false);
	this->keylist[nKeys++] =		this->DefineKey('/', '?', 23, 6, 2, 2, SDL_SCANCODE_SLASH, false);

	this->keylist[nKeys++] =		this->DefineKey(' ', ' ', 8, 8, 12, 2, SDL_SCANCODE_SPACE, false);
	

	// create the keymap of printable characters (lets you find the keys we just made)
	for (i=0;i<128;i++) {
		this->keymap[i] = 255; // initialize all possible inputs to a non-existent key
	}
	for (i=0;i<nKeys;i++) {
		this->keymap[this->keylist[i].keyCode] = i;
		this->keymap[this->keylist[i].shiftCode] = i;
	}

	// backspace -- we use non-printable characters in the keymap so that we can
	// highlight these keys in intro scripts. Some of these are appropriate, some are just
	// whatever C++ escape sequence I could find.
	this->keylist[nKeys++] =		this->DefineKey(' ', ' ', 26, 0, 4, 2, SDL_SCANCODE_BACKSPACE, false);
	this->keymap['\b'] = nKeys-1;	// backspace (\b)
	this->keylist[nKeys++] =		this->DefineKey(' ', ' ', 0, 2, 3, 2, SDL_SCANCODE_TAB, false);
	this->keymap['\t'] = nKeys-1;	// tab (\t)
	this->keylist[nKeys++] =		this->DefineKey(' ', ' ', 0, 4, 4, 2, SDL_SCANCODE_CAPSLOCK, false);
	this->keymap['\f'] = nKeys-1;	// form feed (\f)
	this->keylist[nKeys++] =		this->DefineKey(' ', ' ', 26, 4, 4, 2, SDL_SCANCODE_RETURN, false);
	this->keymap['\n'] = nKeys-1;	// carriage return (\n)
	this->keylist[nKeys++] =		this->DefineKey(' ', ' ', 0, 6, 5, 2, SDL_SCANCODE_LSHIFT, false);
	this->keymap['\a'] = nKeys-1;	// audible bell (\a)
	this->keylist[nKeys++] =		this->DefineKey(' ', ' ', 25, 6, 5, 2, SDL_SCANCODE_RSHIFT, false);
	this->keymap['\v'] = nKeys-1;	// vertical tab (\v)


	// clear the tilemaps
	for (i=0;i<RE_TILEMAP_SIZE_X*RE_TILEMAP_SIZE_Y;i++) {
		this->globals->re.tilemap[0][i] = 1023;
		this->globals->re.tilemap[1][i] = 1023;
		this->globals->re.tilemap[2][i] = 1023;
		this->globals->re.tilemap[3][i] = 1023;
	}

	// draw the blank keyboard (the grey part with no keys)
	for (j = 0; j < 10; j++) {
		for (i = 0; i < 30; i++) {
			this->globals->re.tilemap[2][j*RE_TILEMAP_SIZE_X + i] = re_make_tile(25, 0, false, false, 1);
		}
	}
	for (j = 0; j < 10; j++) {
		this->globals->re.tilemap[2][j*RE_TILEMAP_SIZE_X + (RE_TILEMAP_SIZE_X-1)] = re_make_tile(24, 0, false, false, 1);
		this->globals->re.tilemap[2][j*RE_TILEMAP_SIZE_X + 30] = re_make_tile(24, 0, true, false, 1);
	}
	for (i = 0; i < 30; i++) {
		this->globals->re.tilemap[2][(RE_TILEMAP_SIZE_Y-1)*RE_TILEMAP_SIZE_X + i] = re_make_tile(23, 0, false, false, 1);
		this->globals->re.tilemap[2][10*RE_TILEMAP_SIZE_X + i] = re_make_tile(8, 0, false, false, 1);
	}
	this->globals->re.tilemap[2][RE_TILEMAP_SIZE_X*RE_TILEMAP_SIZE_Y - 1] = re_make_tile(22, 0, false, false, 1);
	this->globals->re.tilemap[2][(RE_TILEMAP_SIZE_Y-1)*RE_TILEMAP_SIZE_X + 30] = re_make_tile(22, 0, true, false, 1);
	this->globals->re.tilemap[2][10*RE_TILEMAP_SIZE_Y + RE_TILEMAP_SIZE_X - 1] = re_make_tile(9, 0, true, false, 1);
	this->globals->re.tilemap[2][10*RE_TILEMAP_SIZE_Y + 30] = re_make_tile(9, 0, false, false, 1);

	// draw the keys
	for (i=0;i<nKeys;i++) {
		this->DrawKey(i, false, false);
	}
	this->keyboardposx = -40;
	this->keyboardposy = -20;

	// draw the "paper"
	for (i = 0; i < RE_TILEMAP_SIZE_X; i++) {
		this->globals->re.tilemap[3][15 * RE_TILEMAP_SIZE_X + i] = re_make_tile(10, 0, false, false, 1);
		for (j = 16; j < 24; j++) {
			this->globals->re.tilemap[3][j*RE_TILEMAP_SIZE_X + i] = re_make_tile(11, 0, false, false, 1);
		}
	}

	// draw the level label
	this->WriteString(LevelLabel, 2, 0, RE_TILEMAP_SIZE_Y -2, 0, 0);

	// setup hblank callback
	this->globals->re.SetHBlank(HBlankCallback, this);
	this->hBlankFrame = 0;

	// copy the intro script
	this->intro = inIntro;
}

/**
 * the main gameplay loop
 */
int CGameplayLoop::Loop() {
	bool GameRunning = true;
	
	SDL_Event evt;
	uint32 keycode = 0;
	uint8 keystate[512];
	uint8 *newkeystate;
	int nKeystate;
	SDL_Keymod modstate;

	bool shift = false;
	bool capslock = false;

	bool fullRedraw = false;

	int i, j, nSprites;

	int cursorPos = 0;
	int currentString = 0;
	int keyPressed = 0;

	int fontOffset = 96;

	uint32 backupPalette[256];
	uint32 fadeoutPalette[256]; // this is so wasteful but I can't think of a better way ...
	uint8 r, g, b;
	Uint64 startTime;
	uint32 errors = 0;
	uint32 nonErrors = 0;

	char tempStr[32];

	bool introLShift = false;
	bool introRShift = false;

	int introIndex = 0;
	char introCommand;
	int introX, introY;
	char introKey;
	char introState = 'r';
	int introTimer;

	// set layer initial state
	this->globals->re.SetLayer(3, true, 0, 0);
	this->globals->re.SetLayer(2, true, keyboardposx, keyboardposy);
	this->globals->re.SetLayer(1, true, keyboardposx+3, keyboardposy+3);
	this->globals->re.SetLayer(0, true, 0, 0);

	// get keyboard initial state
	newkeystate = (uint8 *)SDL_GetKeyboardState(&nKeystate);
	memcpy(keystate, newkeystate, nKeystate*sizeof(uint8));

	// back up the palette
	memcpy(backupPalette, this->globals->re.palette, 256*sizeof(uint32));

	// fade-in loop
	for (i = 0; i < 64; i++) {
		// fade the palette
		for (j = 0; j < 256; j++) {
			r = (backupPalette[j] >> 16) & 0xff;
			g = (backupPalette[j] >> 8) & 0xff;
			b = backupPalette[j] & 0xff;
			r = (r*i)>>6;
			g = (g*i)>>6;
			b = (b*i)>>6;
			this->globals->re.palette[j] = (r<<16) | (g<<8) | b;
		}

		// draw capslock light
		nSprites = nKeys; // we reserve the first nKeys sprites for depressed keys
		this->globals->re.SetSprite(nSprites++, true, 6, 0, 0, 1, 1, 0 - keyboardposx, 32 - keyboardposy, false, false, capslock ? 1 : 0);

		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();
	}
	// restore palette after fade-in
	memcpy(this->globals->re.palette, backupPalette, 256 * sizeof(uint32));

	// intro loop
	GameRunning = true;
	while (GameRunning) {
		while (SDL_PollEvent(&evt)) {
			// if we get the quit signal return immediately
			if (evt.type == SDL_QUIT) {
				FinishGame(this->globals);
				memcpy(this->globals->re.palette, backupPalette, 256 * sizeof(uint32));
				return 0;
			}

			if (evt.type == SDL_KEYDOWN) {
				// if it's escape, return to the menu
				if (evt.key.keysym.sym == SDLK_ESCAPE) {
					memcpy(this->globals->re.palette, backupPalette, 256 * sizeof(uint32));
					return 0;
				}
				// if we were waiting for keypress, move forward
				if (introState == 'w') {
					introState = 'r';
				}
			}
		}


		// while ready to interpret commands
		while (introState == 'r') {
			// get the next one-letter command
			introCommand = intro[introIndex][0];
			if (introCommand == 't') {
				// write text to X/Y coord
				introX = 10*(intro[introIndex][1]-'0') + (intro[introIndex][2]-'0');
				introY = 10*(intro[introIndex][3]-'0') + (intro[introIndex][4]-'0');
				introState = 't';
				introTimer = 5;
			} else if (introCommand == 'e') {
				// end
				introState = 'e';
				GameRunning = false;
			} else if (introCommand == 'w') {
				// wait for keypress
				introState = 'w';
				introIndex++;
			} else if (introCommand == 'p') {
				// pause for a set amount of time
				introState = 'p';
				introTimer = 100*(intro[introIndex][1]-'0') + 10*(intro[introIndex][2]-'0') +  (intro[introIndex][3]-'0');
				introIndex++;
			} else if (introCommand == 'c') {
				// clear text (instantly)
				introX = 10 * (intro[introIndex][1] - '0') + (intro[introIndex][2] - '0');
				introY = 10 * (intro[introIndex][3] - '0') + (intro[introIndex][4] - '0');
				introTimer = 10*(intro[introIndex][5]-'0') + (intro[introIndex][6]-'0');
				while (introTimer > 0) {
					this->globals->re.tilemap[0][introY*RE_TILEMAP_SIZE_X + introX] = re_make_tile(1023, 0, false, false, 1);
					introX++;
					introTimer--;
				}
				introIndex++;
			} else if (introCommand == 'h') {
				// highlight / un-highlight a key
				this->keylist[this->keymap[intro[introIndex][1]]].pressed = (intro[introIndex][2]=='1'?true:false);
				this->DrawKey(this->keymap[intro[introIndex][1]], shift, capslock);
				if (intro[introIndex][1] == '\f' ||
					intro[introIndex][1] == '\a' ||
					intro[introIndex][1] == '\v') {
					// state change for caps lock or shift, full redraw needed
					capslock = this->keylist[this->keymap['\f']].pressed;
					shift = this->keylist[this->keymap['\a']].pressed | this->keylist[this->keymap['\v']].pressed;
					for (i=0;i<this->nKeys;i++) {
						this->DrawKey(i, shift, capslock);
					}
				}
				introIndex++;
			} else {
				// invalid command - exit
				introState = 'e';
				GameRunning = false;
			}

		}

		// if we're rolling out text ...
		if (introState == 't') {
			this->globals->re.tilemap[0][introY*RE_TILEMAP_SIZE_X + introX] = re_make_tile(intro[introIndex][introTimer] + fontOffset, 1, false, false, 1);
			introX++;
			introTimer++;
			if (intro[introIndex][introTimer] == '\0') {
				introState = 'r';
				introIndex++;
			}
		// if we're paused for time
		} else if (introState == 'p') {
			introTimer--;
			if (introTimer <= 0) {
				introState = 'r';
			}
		}

		// draw capslock light
		nSprites = nKeys; // we reserve the first nKeys sprites for depressed keys
		this->globals->re.SetSprite(nSprites++, true, 6, 0, 0, 1, 1, 0 - keyboardposx, 32 - keyboardposy, false, false, capslock ? 1 : 0);

		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();
	}

	// draw the prompt
	this->WriteString("Type these words:", 0, 2, 16, 1, fontOffset);

	// draw the first line of text
	this->WriteString(this->GeneratedLines[0], 0, 2, 18, 1, fontOffset);
	
	// set the cursor start pos
	this->globals->re.tilemap[0][20 * RE_TILEMAP_SIZE_X + 2]  = re_make_tile(7, 1, false, false, 1);

	// get the start time
	startTime = SDL_GetPerformanceCounter();

	GameRunning = true;
	while (GameRunning) {
		while (SDL_PollEvent(&evt)) {
			if (evt.type == SDL_QUIT) {
				return 0;
			}

			// keydown event
			if (evt.type == SDL_KEYDOWN) {
				// is this a printable character?
				if (evt.key.keysym.sym >= 32 && evt.key.keysym.sym <= 127 && this->keymap[evt.key.keysym.sym] != 255) {
					// find out what character got input
					keyPressed = this->keylist[this->keymap[evt.key.keysym.sym]].keyCode;
					if ((evt.key.keysym.mod&(KMOD_LSHIFT | KMOD_RSHIFT)) || ((evt.key.keysym.mod&KMOD_CAPS)&&this->keylist[this->keymap[evt.key.keysym.sym]].capsEffect)) {
						keyPressed = this->keylist[this->keymap[evt.key.keysym.sym]].shiftCode;
					}
					// if this character is the correct character for the string, update cursor pos
					if (this->GeneratedLines[currentString][cursorPos] == keyPressed) {
						this->globals->re.tilemap[0][20 * RE_TILEMAP_SIZE_X + 2 + cursorPos] = re_make_tile(keyPressed+fontOffset, 1, false, false, 1);
						cursorPos++;
						this->globals->re.tilemap[0][20 * RE_TILEMAP_SIZE_X + 2 + cursorPos] = re_make_tile(7, 1, false, false, 1);

						// if we completed a line
						if (this->GeneratedLines[currentString][cursorPos] == '\0') {
							currentString ++;
							cursorPos = 0;
							// if we completed the last string
							if (currentString == 10) {
								// this is a game winning condition TO DO
								GameRunning = false;
							} else {
								// make sure we check to see if there's a string here before overwriting
								this->WriteString("                                                ", 0, 2, 18, 0);
								this->WriteString("                                                ", 0, 2, 20, 0);
								this->WriteString(this->GeneratedLines[currentString], 0, 2, 18, 1, fontOffset);
								this->globals->re.tilemap[0][20 * RE_TILEMAP_SIZE_X + 2 + cursorPos] = re_make_tile(7, 1, false, false, 1);
							}
						}
						// record a correct keystroke
						nonErrors++;
					} else {
						// a printable character was typed that didn't match, so record an error
						errors++;
					}
				} else if (evt.key.keysym.sym == SDLK_ESCAPE) {
					memcpy(this->globals->re.palette, backupPalette, 256 * sizeof(uint32));
					return 0;
				}
			}
		}
		nSprites = 0;

		// get key states if any are different, update key
		newkeystate = (uint8 *)SDL_GetKeyboardState(&nKeystate);
		for (i=0;i<this->nKeys;i++) {
			if (newkeystate[this->keylist[i].scanCode] != keystate[this->keylist[i].scanCode]) {
				this->keylist[i].pressed = newkeystate[this->keylist[i].scanCode];
				this->DrawKey(i, shift, capslock);
			}
		}
		memcpy(keystate, newkeystate, nKeystate);

		// get modifier key states
		modstate = SDL_GetModState();
		fullRedraw = false;

		// check if capslock is on
		if ((modstate & KMOD_CAPS) != capslock) {
			capslock = modstate & KMOD_CAPS;
			fullRedraw = true;
		}
		// check for change in shift key - if change redraw keybaord
		if ((modstate & (KMOD_LSHIFT | KMOD_RSHIFT)) != shift) {
			shift = modstate & (KMOD_LSHIFT | KMOD_RSHIFT);
			fullRedraw = true;
		}
		if (fullRedraw) {
			for (i=0;i<this->nKeys;i++) {
				this->DrawKey(i, shift,capslock);
			}
		}
		// draw capslock light
		nSprites = nKeys; // we reserve the first nKeys sprites for depressed keys
		this->globals->re.SetSprite(nSprites++, true, 6, 0, 0, 1, 1, 0-keyboardposx, 32-keyboardposy, false, false, capslock?1:0);

		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();
	}

	// stop the clock and convert to seconds
	startTime = (SDL_GetPerformanceCounter() - startTime)/SDL_GetPerformanceFrequency();

	// results screen setup
	for (i = 0; i < RE_TILEMAP_SIZE_X*RE_TILEMAP_SIZE_Y; i++) {
		this->globals->re.tilemap[0][i] = re_make_tile(1023, 0, false, false, 1);
	}
	for (j = 0; j < 4; j++) {
		for (i = 0; i < 14; i++) {
			this->globals->re.tilemap[0][(j+6)*RE_TILEMAP_SIZE_X + (i+13)] = re_make_tile(224+(j*16)+i, 2, false, false, 1);
		}
	}
	// results screen fade-out
	for (i = 63; i >= 0; i--) {
		while (SDL_PollEvent(&evt)) {
			// if we get the quit signal return immediately
			if (evt.type == SDL_QUIT) {
				memcpy(this->globals->re.palette, backupPalette, 256 * sizeof(uint32));
				return 0;
			}

		}

		// fade the palette (first 2 palettes only)
		for (j = 0; j < 32; j++) {
			r = (backupPalette[j] >> 16) & 0xff;
			g = (backupPalette[j] >> 8) & 0xff;
			b = backupPalette[j] & 0xff;
			r = (r*(i/2 + 32)) >> 6;
			g = (g*(i/2 + 32)) >> 6;
			b = (b*(i/2 + 32)) >> 6;
			this->globals->re.palette[j] = (r << 16) | (g << 8) | b;
		}

		// scroll top layer
		this->globals->re.SetLayer(0, true, 0, -((160*i)>>6));


		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();
	}
	// draw results
	sprintf(tempStr, "Errors:          %4d", errors);
	this->WriteString(tempStr, 0, 9, 14, 2, 0);
	sprintf(tempStr, "Words Per Minute:%4d", (nonErrors*60)/(5*(int)startTime));
	this->WriteString(tempStr, 0, 9, 12, 2, 0);
	sprintf(tempStr, "Time:           %02d:%02d", (int)(startTime/60), (int)(startTime%60));
	this->WriteString(tempStr, 0, 9, 16, 2, 0);
	
	// display results
	GameRunning = true;
	while (GameRunning) {
		while (SDL_PollEvent(&evt)) {
			// if we get the quit signal return immediately
			if (evt.type == SDL_QUIT) {
				memcpy(this->globals->re.palette, backupPalette, 256 * sizeof(uint32));
				return 0;
			}

			if (evt.type == SDL_KEYDOWN) {
				if (evt.key.keysym.scancode == SDL_SCANCODE_RETURN) {
					GameRunning = false;
				}
			}
		}

		// blinking text
		if ((this->hBlankFrame >> 5) & 0x01) {
			this->WriteString("Press Enter to Continue", 0, 8, 18, 2, 0);
		} else {
			this->WriteString("                       ", 0, 8, 18, 2, 0);
		}

		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();
	}

	// final fadeout
	memcpy(fadeoutPalette, this->globals->re.palette, 256*sizeof(uint32));
	for (i = 64; i >= 0; i--) {
		while (SDL_PollEvent(&evt)) {
			// if we get the quit signal return immediately
			if (evt.type == SDL_QUIT) {
				memcpy(this->globals->re.palette, backupPalette, 256 * sizeof(uint32));
				return 0;
			}

		}

		// fade the palette (first 2 palettes only)
		for (j = 0; j < 256; j++) {
			r = (fadeoutPalette[j] >> 16) & 0xff;
			g = (fadeoutPalette[j] >> 8) & 0xff;
			b = fadeoutPalette[j] & 0xff;
			r = (r*i) >> 6;
			g = (g*i) >> 6;
			b = (b*i) >> 6;
			this->globals->re.palette[j] = (r << 16) | (g << 8) | b;
		}

		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();
	}

	// restore palette before returning to parent function
	memcpy(this->globals->re.palette, backupPalette, 256*sizeof(uint32));

	// remove our hblank callback
	this->globals->re.ClearHBlank();

	return (int) startTime;
}

void CGameplayLoop::DrawKey(int KeyIndex, bool shift, bool caps) {
	int i, j;
	uint16 *tm;
	TypingKey *inKey = this->keylist + KeyIndex;
	uint16 keycode = shift||(caps && inKey->capsEffect )?inKey->shiftCode:inKey->keyCode;

	// draw the key
	tm = this->globals->re.tilemap[2];
	tm += (inKey->y << 6) + inKey->x;
	*tm = re_make_tile(inKey->pressed?3:0, 0, false, false, 1);
	*(tm+64) = re_make_tile(inKey->pressed?19:16, 0, false, false, 1);
	tm++;
	for (i=1;i<inKey->w-1;i++) {
		*tm = re_make_tile(inKey->pressed?4:1, 0, false, false, 1);
		*(tm+64) = re_make_tile(inKey->pressed?20:17, 0, false, false, 1);
		tm++;
	}
	*tm = re_make_tile(inKey->pressed?5:2, 0, false, false, 1);
	*(tm+64) = re_make_tile(inKey->pressed?21:18, 0, false, false, 1);

	// draw the letter
	tm = this->globals->re.tilemap[1];
	tm += (inKey->y << 6) + inKey->x;
	tm += 64;
	tm += inKey->w-1;
	if (inKey->pressed) {
		*tm = re_make_tile(1023, 0, false, false, 1);
		this->globals->re.SetSprite(KeyIndex, true, keycode, 0, 0, 1, 1, (inKey->x+inKey->w)*8 - this->keyboardposx - 11, (inKey->y+inKey->h)*8 - this->keyboardposy - 10, false, false, 0);
	} else {
		*tm = re_make_tile(keycode, 0, false, false, 1);
		this->globals->re.SetSprite(KeyIndex, false, 1023, 0, 0, 1, 1, inKey->x * 8 - this->keyboardposx, inKey->y * 8 - this->keyboardposy, false, false, 0);
	}
	
}

void CGameplayLoop::HBlank(CRetroEngine *re, int scanline) {
	if (scanline == 0) {
		this->hBlankFrame++;
		re->SetLayer(3, true, 0, 0);
	}
	// paper starts at line 15
	if (scanline == 15 * 8) {
		re->SetLayer(3, true, this->hBlankFrame, 0);
	}
};

/**
 * We can't set a member function as our HBlank callback, so we use this
 * workaround where *payload is used to reference the class.
 */
void HBlankCallback(CRetroEngine *re, int scanline, void *payload) {
	CGameplayLoop * loop = (CGameplayLoop *)payload;
	loop->HBlank(re, scanline);
}
