#pragma once
#ifndef _GAMEPLAYLOOP_H
#define _GAMEPLAYLOOP_H

#include "GlobalDataStruct.h"
#include "RetroEngine.h"

/**
 * Descriptor for a key displayed on the screen.
 */
struct TypingKey {
	uint8 keyCode;				// ASCII code 
	uint8 shiftCode;			// ASCII code when shift is pressed
	uint16 x, y, w, h; 			// key size and position in tiles
	bool pressed;				// current state
	SDL_Scancode scanCode;		// scan code to associate with this key in event handler
	bool capsEffect;			// whether to display shiftCode when CAPS LOCK is active
};

/**
 * Class describing the main gameplay loop state
 */
class CGameplayLoop {
private:
	char GeneratedLines[10][37];	// 36 character line + \0
	TypingKey keylist[64];			// more than enough for a keyboard
	uint8 keymap[128];				// maps ascii codes to keylist
	GlobalData *globals;			// global data reference to keep
	int nKeys;						// number of keys to draw
	int hBlankFrame;				// framenumber for hblank callback
	char **intro;					// the intro script

	int keyboardposx, keyboardposy;

	/**
	 * convenience function to define a key all in one line
	 */
	TypingKey DefineKey(uint8 inKC, uint8 inSC, uint16 inX, uint16 inY, uint16 inW, uint16 inH, SDL_Scancode inScancode, bool inCapsEffect) {
		TypingKey tmp;
		tmp.keyCode = inKC;
		tmp.shiftCode = inSC;
		tmp.x = inX;
		tmp.y = inY;
		tmp.w = inW;
		tmp.h = inH;
		tmp.pressed = false;
		tmp.scanCode = inScancode;
		tmp.capsEffect = inCapsEffect;

		return tmp;
	};

	void DrawKey(int KeyIndex, bool shift, bool caps);
	
	/**
	 * Writes sting of text to a background layer. Use offset to choose second font.
	 */
	void WriteString(const char *str, int layer, int x, int y, int pal, int offset = 0) {
		re_tile *writeptr = this->globals->re.tilemap[layer] + (y*RE_TILEMAP_SIZE_X + x);
		while (*str != '\0') {
			*writeptr = re_make_tile(*str + offset, pal, false, false, 1);
			writeptr++;
			str++;
		}
	};

public:
	void Init(GlobalData *inGlobals, char **wordList, int nWords, char *LevelLabel, char **inIntro);
	int Loop();
	void HBlank(CRetroEngine *re, int scanline);
};

/**
 * Callback function so that we can do hblank effect
 */
void HBlankCallback(CRetroEngine *re, int scanline, void *payload);

#endif // _GAMEPLAYLOOP_H
