/*
	Global Data Structures

	Defines data types that are used in all sorts of places.

	ONLY PUT DECLARATIONS IN THIS FILE

 */
#pragma once
#ifndef _GLOBALDATASTRUCT_H
#define _GLOBALDATASTRUCT_H

#include "SDL2/SDL.h"
#include "RetroEngine.h"
#include "sdl_virtual_screen.h"

#ifdef WIN32
// create a usleep() function that matches the unix standard
// so that we don't need to adjust code elsewhere
#include <Windows.h>
#define usleep(a) Sleep((a)/1000 + 1)
#else
// include unistd.h for usleep()
#include <unistd.h>
#endif

/**
 * A way to simulate waiting for the old notion of waiting for a 
 * vsync interrupt. Designed to release the CPU as much as possible
 * while waiting.
 */
class CPacemaker {
public:
	Uint64 interval;
	void wait() {
		Uint64 startBeat = SDL_GetPerformanceCounter()/this->interval;
		do {
			usleep(999);
		} while (startBeat == SDL_GetPerformanceCounter() / this->interval);
	}
};

/**
 * GlobalData holds things that /would be/ global in an old C-style program,
 * but instead we instantiate this once inside main(), then pass its pointer
 * where we need it. We get the benefits of globally-available data without
 * the risks of unscoped data.
 */
typedef struct {
	SDL_Window *win;
	CRetroEngine re;
	CSDLVirtualScreen screen;
	CPacemaker pacemaker;
} GlobalData;

/**
 * LevelDataStruct describes the type of information needed to define a level.
 * Used by level1.cpp and MainMenu.cpp.
 */
typedef struct {
	char *levelWords[255];			// up to 255 words per level
	char levelLen;					// how many words are actually in the array levelWords
	char *levelLabel;				// string label to put in menu
	char *levelIntro[255];			// up to 255 lines of introductory text/scripting events
} LevelDataStruct;


#endif //_GLOBALDATASTRUCT_H