#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "MainMenu.h"
#include "sdl_virtual_screen.h"
#include "level1.h"
#include "GameplayLoop.h"

#ifdef WIN32
#include <Windows.h>
#include <Shlobj.h>
#endif
#ifdef __linux__
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#endif

extern void FinishGame(GlobalData *globals);

void CMainMenu::Init(GlobalData *inGlobals) {
	int i;

	this->globals = inGlobals;

	// for testing purposes let's initialize a menu with 20 items
	if (this->menuItems != NULL) {
		free(this->menuItems);
	}

	this->nMenuItems = 10;
	this->menuItems = (MainMenuEntry *)malloc(20*sizeof(MainMenuEntry));
	for (i=0;i<this->nMenuItems;i++) {
		this->menuItems[i].bestTime = -1;
	}

	this->menuItems[0].data = &Data_Level01;
	this->menuItems[0].label = Data_Level01.levelLabel;

	this->menuItems[1].data = &Data_Level02;
	this->menuItems[1].label = Data_Level02.levelLabel;

	this->menuItems[2].data = &Data_Level03;
	this->menuItems[2].label = Data_Level03.levelLabel;

	this->menuItems[3].data = &Data_Level04;
	this->menuItems[3].label = Data_Level04.levelLabel;

	this->menuItems[4].data = &Data_Level05;
	this->menuItems[4].label = Data_Level05.levelLabel;

	this->menuItems[5].data = &Data_Level06;
	this->menuItems[5].label = Data_Level06.levelLabel;

	this->menuItems[6].data = &Data_Level07;
	this->menuItems[6].label = Data_Level07.levelLabel;

	this->menuItems[7].data = &Data_Level08;
	this->menuItems[7].label = Data_Level08.levelLabel;

	this->menuItems[8].data = &Data_Level09;
	this->menuItems[8].label = Data_Level09.levelLabel;

	this->menuItems[9].data = &Data_Level10;
	this->menuItems[9].label = Data_Level10.levelLabel;

	currentMenuItem = 0;
	this->loadBestTimes();

	this->DrawMenu();

}

void CMainMenu::fadeIn() {
	int i, j, r, g, b;
	uint32 backupPalette[256];

	// store palette
	memcpy(backupPalette, this->globals->re.palette, 256*sizeof(uint32));
	
	// fade-in loop
	for (i = 0; i < 64; i++) {
		// fade the palette
		for (j = 0; j < 256; j++) {
			r = (backupPalette[j] >> 16) & 0xff;
			g = (backupPalette[j] >> 8) & 0xff;
			b = backupPalette[j] & 0xff;
			r = (r*i)>>6;
			g = (g*i)>>6;
			b = (b*i)>>6;
			this->globals->re.palette[j] = (r<<16) | (g<<8) | b;
		}
		
		this->globals->re.SetLayer(2, true, i & 0x01, 0);

		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();
	}
	
	// restore palette after fade-in
	memcpy(this->globals->re.palette, backupPalette, 256 * sizeof(uint32));
}

void CMainMenu::fadeOut() {
	int i, j, r, g, b;
	uint32 backupPalette[256];

	// store palette
	memcpy(backupPalette, this->globals->re.palette, 256*sizeof(uint32));
	
	// fade-out loop
	for (i = 63; i >= 0; i--) {
		// fade the palette
		for (j = 0; j < 256; j++) {
			r = (backupPalette[j] >> 16) & 0xff;
			g = (backupPalette[j] >> 8) & 0xff;
			b = backupPalette[j] & 0xff;
			r = (r*i)>>6;
			g = (g*i)>>6;
			b = (b*i)>>6;
			this->globals->re.palette[j] = (r<<16) | (g<<8) | b;
		}

		this->globals->re.SetLayer(2, true, i & 0x01, 0);

		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();
	}
	
	// restore palette after fade-in
	memcpy(this->globals->re.palette, backupPalette, 256 * sizeof(uint32));
}

int CMainMenu::Loop() {
	bool GameRunning;
	SDL_Event evt;

	int i;
	
	CGameplayLoop gameLoop;
	LevelDataStruct *d;

	int newBest;
	int timer = 0;
	
	this->fadeIn();

	// main loop
	GameRunning = true;
	while (GameRunning) {
		while (SDL_PollEvent(&evt)) {
			// if we get the quit signal return immediately
			if (evt.type == SDL_QUIT) {
				FinishGame(this->globals);
			}

			if (evt.type == SDL_KEYDOWN) {
				if (evt.key.keysym.sym == SDLK_DOWN) {
					if (this->currentMenuItem < this->nMenuItems-1) {
						// advance to next
						this->currentMenuItem++;
						this->DrawMenu();
						// scroll down
						for (i=0;i<16;i++) {
							timer++;
							this->globals->re.SetLayer(2, true, timer &0x01, 0);
							this->globals->re.SetLayer(3, true, timer % (8*6), 0);
							this->globals->re.SetLayer(1, true, 0, i*2);
							this->globals->re.Draw();
							this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
							this->globals->pacemaker.wait();
							this->globals->screen.RenderTextureToScreen();        
						}
					}
				} else if (evt.key.keysym.sym == SDLK_UP) {
					if (this->currentMenuItem > 0) {
						// move back to previous
						this->currentMenuItem--;
						this->DrawMenu();
						// scroll up
						for (i=0;i<16;i++) {
							timer++;
							this->globals->re.SetLayer(2, true, timer &0x01, 0);
							this->globals->re.SetLayer(3, true, timer % (8*6), 0);
							this->globals->re.SetLayer(1, true, 0, 64-i*2);
							this->globals->re.Draw();
							this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
							this->globals->pacemaker.wait();
							this->globals->screen.RenderTextureToScreen();        
						}
					}
				} else if (evt.key.keysym.sym == SDLK_RETURN) {
					if (menuItems[currentMenuItem].data != NULL) {
						d = menuItems[currentMenuItem].data;
						this->fadeOut();
						gameLoop.Init(globals, d->levelWords, d->levelLen, d->levelLabel, d->levelIntro);
						newBest = gameLoop.Loop();
						if (newBest > 0) {
							if ((menuItems[currentMenuItem].bestTime < 0) || (newBest < menuItems[currentMenuItem].bestTime)) {
								menuItems[currentMenuItem].bestTime = newBest;
								this->saveBestTimes();
							}
						}
						this->DrawMenu();
						this->fadeIn();
					}
				} else if (evt.key.keysym.sym == SDLK_ESCAPE) {
					GameRunning = false;
				}
			}
		}

		timer++;
		this->globals->re.SetLayer(2, true, timer &0x01, 0);
		this->globals->re.SetLayer(3, true, timer % (8*6), 0);

		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();        
	}
	this->fadeOut();
	return 1;
}

void CMainMenu::DrawMenu() {
	int x, y, i, j, pal;
	char bestTime[32];
	int fontOffset = 96;

	// turn off unused sprites
	for (i=0;i<RE_NUM_SPRITES;i++)  this->globals->re.ClearSprite(i);

	// turn off any previously set hblank callbacks
	this->globals->re.ClearHBlank();

	// turn off unused layers
	this->globals->re.SetLayer(0, false, 0, 0);
	this->globals->re.SetLayer(1, false, 0, 0);
	this->globals->re.SetLayer(2, false, 0, 0);
	this->globals->re.SetLayer(3, false, 0, 0);

	// clear tilemaps
	for (y=0;y<RE_TILEMAP_SIZE_Y;y++) {
		for (x=0; x<RE_TILEMAP_SIZE_X;x++) {
			this->globals->re.tilemap[1][y*RE_TILEMAP_SIZE_X+x] = re_make_tile(' ', 0, 0, 0, 0);
			this->globals->re.tilemap[2][y*RE_TILEMAP_SIZE_X+x] = re_make_tile(' ', 0, 0, 0, 0);
			this->globals->re.tilemap[3][y*RE_TILEMAP_SIZE_X+x] = re_make_tile(' ', 0, false, false, 0);
		}
	}

	// draw the sky on layer 3
	this->globals->re.SetLayer(3, true, 0, 0);
	for (y=0;y<20;y++) {
		for (x=0;x<RE_TILEMAP_SIZE_X; x++) {
			this->globals->re.tilemap[3][y*RE_TILEMAP_SIZE_X + x] = re_make_tile((0x120+(y>>1)), 3, false, false, 0);
		}
	}
	for (x=0;x<RE_TILEMAP_SIZE_X; x++) {
		this->globals->re.tilemap[3][20*RE_TILEMAP_SIZE_X + x] = re_make_tile((0x12A+(x%6)), 3, false, false, 0);
		this->globals->re.tilemap[3][21*RE_TILEMAP_SIZE_X + x] = re_make_tile((0x13A+(x%6)), 3, false, false, 0);
		this->globals->re.tilemap[3][22*RE_TILEMAP_SIZE_X + x] = re_make_tile(0x130, 3, false, false, 0);
	}

	this->globals->re.SetLayer(2, true, 0, 0);

	// draw the "screen" on layer 2
	/*
	for (y=0;y<23;y++) {
		for (x=2; x<36; x++) {
			this->globals->re.tilemap[2][y*RE_TILEMAP_SIZE_X+x] = re_make_tile(7, 0, false, false, 0);
		}
	}
	*/

	// draw the "selection" bar on layer 2
	for (y=3;y<7;y++) {
		for (x=0; x<RE_TILEMAP_SIZE_X;x++) {
			this->globals->re.tilemap[2][y*RE_TILEMAP_SIZE_X+x] = re_make_tile(11, 0, 0, 0, 0);
		}
	}
	y = 3;
	for (x=0; x<RE_TILEMAP_SIZE_X;x++) {
		this->globals->re.tilemap[2][y*RE_TILEMAP_SIZE_X+x] = re_make_tile(10, 0, 0, 0, 0);
	}
	y = 6;
	for (x=0; x<RE_TILEMAP_SIZE_X;x++) {
		this->globals->re.tilemap[2][y*RE_TILEMAP_SIZE_X+x] = re_make_tile(10, 0, 0, true, 0);
	}

	// we're going to draw the menu to layer 1
	this->globals->re.SetLayer(1, true, 0, 32); // start with 32 px off the top of the screen so that it can scroll into view

	y = 0;
	for (i = currentMenuItem-2; i<currentMenuItem+9; i++) {
		if (i>=0 && i<nMenuItems) {
			// figure out indent
			x = (i == currentMenuItem)?4:4;
			pal = (i == currentMenuItem)?0:1;
			// write the first line of text
			j = 0;
			while (menuItems[i].label[j] != '\0') {
				this->globals->re.tilemap[1][y*RE_TILEMAP_SIZE_Y+x] = re_make_tile(menuItems[i].label[j], pal, 0, 0, 0);
				x++;
				j++;
			}
			y++;
			// write the best time;
			x = (i == currentMenuItem)?4:4;
			x += 4;
			if (menuItems[i].bestTime < 0) {
				sprintf(bestTime, "Best Time: --:--");
			} else {
				sprintf(bestTime, "Best Time: %d:%02d", menuItems[i].bestTime/60, menuItems[i].bestTime%60);
			}
			j = 0;
			while (bestTime[j] != '\0') {
				this->globals->re.tilemap[1][y*RE_TILEMAP_SIZE_Y+x] = re_make_tile(bestTime[j], pal, 0, 0, 0);
				x++;
				j++;
			}
			y++;
			// advance 2 lines
			y+= 2;
		} else {
			// blank item, advance 4 lines
			y += 4;
		}
	}
}

void CMainMenu::saveBestTimes() {
	FILE *fp = NULL;
	uint32 tempWord;
	int i;

#ifdef WIN32
	WCHAR path[MAX_PATH];
	if (SUCCEEDED(SHGetFolderPathW(NULL, CSIDL_APPDATA, NULL, 0, path))) {
		wcscat(path, L"\\typingAcademyTimes.bin");
		fp = _wfopen(path, L"wb");
		if (!fp) return;

	} else {
		return;
	}
#endif

#ifdef __linux__
	struct passwd *pw = getpwuid(getuid());
	const char *homedir = pw->pw_dir;
	char path[256] = "";
	strcat(path, homedir);
	strcat(path, "/.typeAcademyTimes.bin");
	// printf("using filename: %s\n", path);
	fp = fopen(path, "wb");
	if (!fp) return;
#endif

	if (!fp) return;  // we should only reach here if we're compiled in NOT linux OR Windows (OS X?)
	
	for (i=0;i<this->nMenuItems;i++) {
		if (this->menuItems[i].bestTime > 0) {
			tempWord = this->menuItems[i].bestTime & 0x7FFFFFFF;
			tempWord |= 0x80000000;
			fwrite(&tempWord, 4, 1, fp);
		} else {
			tempWord = 0x00000000;
			fwrite(&tempWord, 4, 1, fp);
		}
	}

	fclose(fp);
}

void CMainMenu::loadBestTimes() {
	FILE *fp = NULL;
	uint32 tempWord;
	int i;

#ifdef WIN32
	WCHAR path[MAX_PATH];
	if (SUCCEEDED(SHGetFolderPathW(NULL, CSIDL_APPDATA, NULL, 0, path))) {
		wcscat(path, L"\\typingAcademyTimes.bin");
		fp = _wfopen(path, L"rb");
		if (!fp) return;

	} else {
		return;
	}
#endif

#ifdef __linux__
	struct passwd *pw = getpwuid(getuid());
	const char *homedir = pw->pw_dir;
	char path[256] = "";
	strcat(path, homedir);
	strcat(path, "/.typeAcademyTimes.bin");
	// printf("using filename: %s\n", path);
	fp = fopen(path, "rb");
	if (!fp) return;
#endif	

	if (!fp) return;  // we should only reach here if we're compiled in NOT linux OR Windows (OS X?)

	i = 0;
	while ((i < this->nMenuItems) && (1 == fread(&tempWord, 4, 1, fp))) {
		// if this is a valid time
		if (tempWord & 0x80000000) {
			tempWord = tempWord & 0x7FFFFFFF;
			this->menuItems[i].bestTime = tempWord;
		// if it's not a valid time
		} else {
			this->menuItems[i].bestTime = -1;
		}
		i++;
	}

	fclose(fp);
}
