#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "TitleScreen.h"
#include "sdl_virtual_screen.h"


extern void FinishGame(GlobalData *globals);


void CTitleScreen::Init(GlobalData *inGlobals) {
	this->globals = inGlobals;

	this->DrawTitleScreen();
}

void CTitleScreen::fadeIn() {
    int i, j, r, g, b;
    uint32 backupPalette[256];

    // store palette
    memcpy(backupPalette, this->globals->re.palette, 256*sizeof(uint32));
    
	// fade-in loop
	for (i = 0; i < 64; i++) {
		// fade the palette
		for (j = 0; j < 256; j++) {
			r = (backupPalette[j] >> 16) & 0xff;
			g = (backupPalette[j] >> 8) & 0xff;
			b = backupPalette[j] & 0xff;
			r = (r*i)>>6;
			g = (g*i)>>6;
			b = (b*i)>>6;
			this->globals->re.palette[j] = (r<<16) | (g<<8) | b;
		}
        
        this->globals->re.SetLayer(2, true, i & 0x01, 0);

		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();
	}
    
	// restore palette after fade-in
	memcpy(this->globals->re.palette, backupPalette, 256 * sizeof(uint32));
}

void CTitleScreen::fadeOut() {
    int i, j, r, g, b;
    uint32 backupPalette[256];

    // store palette
    memcpy(backupPalette, this->globals->re.palette, 256*sizeof(uint32));
    
	// fade-out loop
	for (i = 63; i >= 0; i--) {
		// fade the palette
		for (j = 0; j < 256; j++) {
			r = (backupPalette[j] >> 16) & 0xff;
			g = (backupPalette[j] >> 8) & 0xff;
			b = backupPalette[j] & 0xff;
			r = (r*i)>>6;
			g = (g*i)>>6;
			b = (b*i)>>6;
			this->globals->re.palette[j] = (r<<16) | (g<<8) | b;
		}

        this->globals->re.SetLayer(2, true, i & 0x01, 0);

		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();
	}
    
	// restore palette after fade-in
	memcpy(this->globals->re.palette, backupPalette, 256 * sizeof(uint32));
}

int CTitleScreen::Loop() {
	bool GameRunning = true;
	unsigned int timer = 0;
	int opacity = 0;
	SDL_Event evt;

	this->fadeIn();

	while (GameRunning) {
        while (SDL_PollEvent(&evt)) {
			// if we get the quit signal return immediately
			if (evt.type == SDL_QUIT) {
				FinishGame(this->globals);
			}

            if (evt.type == SDL_KEYDOWN) {
				if (evt.key.keysym.sym == SDLK_ESCAPE) {
                    FinishGame(this->globals);
                } else {
					GameRunning = false;
					timer += 127; // skip past fade-in
				}
            }
		}

        timer++;

		// scroll the "ground" layer
		this->globals->re.SetLayer(3, true, timer % (8*6), 0);

		// fade in the logo
		opacity = (timer>127?127:timer);
		this->globals->re.SetLayer(0, true, -96, -48, RE_BLEND_BASIC, opacity);

		// blink "press start"
		this->globals->re.SetLayer(1, ((timer>>5)&0x01), 0, 0, RE_BLEND_OFF, 127);

		this->globals->re.Draw();
		this->globals->screen.DirectUpdate(this->globals->re.GetScreen());
		this->globals->pacemaker.wait();
		this->globals->screen.RenderTextureToScreen();       
	}

	this->fadeOut();

	return 1;
}

void CTitleScreen::DrawTitleScreen() {
	int x, y, offs;
	char *pressStart = "Press Any Key";
	int pressStartLen = strlen(pressStart);

	// grab a pointer to save my sanity typing this stuff out
	CRetroEngine *re = &(this->globals->re);
	
	// initialize all layer states
	re->SetLayer(0, false, 0, 0, RE_BLEND_OFF, 255);
	re->SetLayer(1, false, 0, 0, RE_BLEND_OFF, 255);
	re->SetLayer(2, false, 0, 0, RE_BLEND_OFF, 255);
	re->SetLayer(3, false, 0, 0, RE_BLEND_OFF, 255);

	// clear tilemaps
    for (y=0;y<RE_TILEMAP_SIZE_Y;y++) {
        for (x=0; x<RE_TILEMAP_SIZE_X;x++) {
            re->tilemap[0][y*RE_TILEMAP_SIZE_X+x] = re_make_tile(' ', 0, false, false, 0);
            re->tilemap[1][y*RE_TILEMAP_SIZE_X+x] = re_make_tile(' ', 0, false, false, 0);
			re->tilemap[2][y*RE_TILEMAP_SIZE_X+x] = re_make_tile(' ', 0, false, false, 0);
            re->tilemap[3][y*RE_TILEMAP_SIZE_X+x] = re_make_tile(' ', 0, false, false, 0);
        }
    }

    // draw the sky on layer 3
    re->SetLayer(3, true, 0, 0);
    for (y=0;y<20;y++) {
        for (x=0;x<RE_TILEMAP_SIZE_X; x++) {
            re->tilemap[3][y*RE_TILEMAP_SIZE_X + x] = re_make_tile((0x120+(y>>1)), 3, false, false, 0);
        }
    }
    for (x=0;x<RE_TILEMAP_SIZE_X; x++) {
        re->tilemap[3][20*RE_TILEMAP_SIZE_X + x] = re_make_tile((0x12A+(x%6)), 3, false, false, 0);
        re->tilemap[3][21*RE_TILEMAP_SIZE_X + x] = re_make_tile((0x13A+(x%6)), 3, false, false, 0);
        re->tilemap[3][22*RE_TILEMAP_SIZE_X + x] = re_make_tile(0x130, 3, false, false, 0);
    }

	// draw the logo on layer 0
	re->SetLayer(0, true, 128, 0, RE_BLEND_BASIC, 0);
	offs = 0x140;
	for (y=0; y<6; y++) {
		for (x=0; x<16; x++) {
			re->tilemap[0][y*RE_TILEMAP_SIZE_X + x] = re_make_tile(offs, 4, false, false, 0);
			++offs;
		}
	}

	// draw "press start" on layer 1
	y = 16;				// draw on 16th row
	x = (320/8);		// temporariliy store screen width in tiles
	x = (x - pressStartLen) >> 1;		// start in such a way as to center text
	offs = 0;			// move to start of string
	while (pressStart[offs] != '\0') {
		re->tilemap[1][y*RE_TILEMAP_SIZE_X + x] = re_make_tile(pressStart[offs], 0, false, false, 0);
		offs++;
		x++;
	}


}