#ifndef _SDL_VIRTUAL_SCREEN_H
#define _SDL_VIRTUAL_SCREEN_H

#include "SDL2/SDL.h"						// SDL library

/*
	Outlines a basic virtual screen that will allow you to send
	raw RGB out to SDL with minimal fuss.
*/

// Set some hard limits on width & height so that
// we can avoid an initialization SNAFU. This prevents
// the program from trying to allocate a ton of
// memory when you give it garbage values and instead
// returns an error message safely.
const int SDL_VSCR_MAXWIDTH = 1024;
const int SDL_VSCR_MAXHEIGHT = 1024;

class CSDLVirtualScreen {
	// member variables
	private:
		SDL_Texture *tex;
		SDL_Window *win;
		SDL_Renderer *ren;
		int width;
		int height;

	// construct and destruct
	public:
		CSDLVirtualScreen():
			tex(nullptr),
			win(nullptr),
			ren(nullptr),
			width(0),
			height(0) {
		};
		~CSDLVirtualScreen();
	// functions
		int Initialize(int inWidth, int inHeight, SDL_Window *inWin);
		int Update(unsigned char *inData);
		int Update(float *inData);
		int DirectUpdate(unsigned char *inData);
		inline SDL_Renderer *GetRenderer() {return this->ren;}

		void RenderTextureToScreen();
};




#endif